from elasticsearch import Elasticsearch
import os
import zipfile
import shutil
import urllib.request
import logging
import lzma
import json


def create_index_from_json(index_name, file_path, max_docs=None):
    """Create an index from json file formats.
    Read each file line by line, parse each line as json
        jsonl.xz
        json  : must be a json file containing a list

    Arguments:
        file_path {str} -- path to case.law bulk file
    Keyword Arguments:
        max_docs {int} -- maximum size of records to use in creating index.
        small default can be used to enable quick testing (e.g: {2000}).
        set this to None to use the entire data file.
    """

    es = Elasticsearch([{'host': 'localhost', 'port': 9200}])

    index_settings = {
        "settings": {
            "analysis": {
                "analyzer": {
                    "stop_analyzer": {
                        "type": "standard",
                        "stopwords": "_english_"
                    }
                }
            }
        },
        "mappings": {
            "properties": {
                "casebody.data.opinions.text": {
                    "type": "text",
                    "analyzer": "stop_analyzer"
                },
                "name": {
                    "type": "text",
                    "analyzer": "stop_analyzer"
                }
            }
        }
    }
    es.indices.create(
        index=index_name, body=index_settings, ignore=400)

    extension = os.path.splitext(file_path)[1]
    logging.info(">> Creating index using file " + file_path)
    i = 0
    if extension == ".xz":
        with lzma.open(file_path) as f:
            for line in f:
                i += 1
                line = json.loads(str(line, 'utf8'))
                try:
                    index_status = es.index(
                        index=index_name, id=i, body=line)
                except Exception as e:
                    logging.info(
                        "An error has occurred while creating index " + str(e))
                    break
                # logging.info(index_status)
                if (i > max_docs):
                    break
    elif extension == ".json":
        with open(file_path) as f:
            data = json.load(f)
            for line in data:
                try:
                    index_status = es.index(
                        index=index_name, id=i, body=line, op_type="create")
                except Exception as e:
                    logging.info(
                        "An error has occurred while creating index " + str(e))
                    break
                if (i > max_docs):
                    break
    logging.info(">> Creating index complete ")


def download_data(data_url, source_name):
    """Download Zip datafile from case.law
    Arguments:
        data_url {str} -- url path dataset
        source_name {str} -- name for dataset
    """
    # create data directory
    os.makedirs("data", exist_ok=True)
    # download data from caselaw
    zip_file_path = source_name + ".zip"
    logging.info("Downloading data file for " + source_name)
    urllib.request.urlretrieve(data_url, zip_file_path)
    logging.info(">> Downloaded data file " + zip_file_path)

    extract_dir = "temp" + source_name
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        zip_ref.extractall(extract_dir)
    data_file = os.path.join(extract_dir, os.listdir(
        extract_dir)[0], "data", "data.jsonl.xz")
    final_file_path = os.path.join("data", source_name + "jsonl.xz")
    shutil.copyfile(data_file, final_file_path)
    logging.info(">> Extracted and moved jsonl file to data folder")
    shutil.rmtree(extract_dir)
    os.remove(zip_file_path)
    return final_file_path


def import_case_data(max_docs=2000):
    """Download new mexico legal case documents,
    import the first 2,000 cases.
    """
    caselaw_data_paths = [
        ["https://api.case.law/v1/bulk/22411/download/", "newmexico"]
    ]
    for data_path in caselaw_data_paths:
        file_path = download_data(data_path[0], data_path[1])
        create_index_from_json("cases", file_path, max_docs=max_docs)



from neuralqa.retriever import Retriever
from elasticsearch import Elasticsearch, ConnectionError
import logging
logging.getLogger("elasticsearch").setLevel(logging.CRITICAL)


class ElasticSearchRetriever(Retriever):
    def __init__(self, index_type="elasticsearch", host="localhost", port=9200):
        Retriever.__init__(self, index_type)

        self.es = Elasticsearch([{'host': host, 'port': port}])
        self.isAvailable = self.es.ping()

    def run_query(self, index_name, search_query):
        """Makes a query to the elastic search server with the given search_query parameters.
        Also returns opinion_excerpt script field, which is a substring of the first opinion in the case

        Arguments:
            search_query {[dictionary]} -- [contains a dictionary that corresponds to an elastic search query on the ]

        Returns:
            [dictionary] -- [dictionary of results from elastic search.]
        """
        query_result = None

        # return error as result on error.
        # Calling function should check status before parsing result
        try:
            query_result = {
                "status": True,
                "result": self.es.search(index=index_name, body=search_query)
            }
        except ConnectionRefusedError as e:
            query_result = {
                "status": False,
                "result": str(e)
            }
        except Exception as e:
            query_result = {
                "status": False,
                "result": str(e)
            }
        return query_result

    def test_connection(self):
        try:
            self.es.cluster.health()
            return True
        except ConnectionError:
            return False
        except Exception as e:
            print('An unknown error occured connecting to ElasticSearch: %s' % e)
            return False





from neuralqa.reader import BERTReader, ReaderPool
from neuralqa.server.routehandlers import Handler
from neuralqa.retriever import ElasticSearchRetriever
from neuralqa.utils import ConfigParser


import os
import logging
import time
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles


config_path = os.environ.get("NEURALQA_CONFIG_PATH")
app_config = ConfigParser(config_path)

app = FastAPI()
api = FastAPI(root_path="/api")

root_file_path = os.path.dirname(os.path.abspath(__file__))
static_folder_root = os.path.join(root_file_path, "ui/build")

app.mount("/api", api)
app.mount("/", StaticFiles(directory=static_folder_root, html=True),
          name="ui")


@api.get('/config')
async def ui_config():
    config = app_config.config["ui"]
    # show only listed models to ui
    config["queryview"]["options"]["relsnip"] = app_config.config["relsnip"]
    config["queryview"]["options"]["samples"] = app_config.config["samples"]
    config["queryview"]["options"]["expander"] = app_config.config["expander"]
    config["queryview"]["options"]["reader"] = app_config.config["reader"]
    config["queryview"]["options"]["retriever"] = app_config.config["retriever"]
    return config

# # Define a Reader Pool
reader_pool = ReaderPool(app_config.config["reader"])

# # define the search index to be used if any
retriever = ElasticSearchRetriever()

handlers = Handler(reader_pool, retriever)
# handlers = Handler(None, None)
api.include_router(handlers.router)





from neuralqa.utils import ConfigParser
import time
from fastapi import APIRouter
from typing import Optional
from neuralqa.server.routemodels import Document, Answer, Explanation


class Handler:
    def __init__(self, reader_pool, retriever):
        router = APIRouter()
        self.router = router

        self._reader_pool = reader_pool
        self._retriever = retriever

        @router.post("/answers")
        async def get_answers(params: Answer):

            """Generate an answer for the given search query.
            Performed as two stage process
            1.) Get sample passages from neighbourhood provided by matches by elastic search
            2.) Used BERT Model to identify exact answer spans

            Returns:
                [type] -- [description]
            """

            # switch to the selected model
            self._reader_pool.selected_model = params.reader

            included_fields = ["name"]
            search_query = {
                "_source": included_fields,
                "query": {
                    "multi_match": {
                        "query":    params.question,
                        "fields": ["casebody.data.opinions.text", "name"]
                    }
                },
                "highlight": {
                    "fragment_size": params.highlight_span,
                    "fields": {
                        "casebody.data.opinions.text": {"pre_tags": [""], "post_tags": [""]},
                        "name": {}
                    }
                },
                "size": params.max_documents
            }

            answer_holder = []
            response = {}
            start_time = time.time()

            # answer question based on provided context
            if (params.retriever == "manual"):
                answers = self._reader_pool.model.answer_question(
                    params.question, params.context, stride=params.tokenstride)
                for answer in answers:
                    answer["index"] = 0
                    answer_holder.append(answer)
            # answer question based on retrieved passages from elastic search
            else:
                query_result = self._retriever.run_query(
                    params.retriever, search_query)
                if query_result["status"]:
                    query_result = query_result["result"]
                    for i, hit in enumerate(query_result["hits"]["hits"]):
                        if ("casebody.data.opinions.text" in hit["highlight"]):
                            # context passage is a concatenation of highlights
                            if (params.relsnip):
                                context = " .. ".join(
                                    hit["highlight"]["casebody.data.opinions.text"])
                            else:
                                context = str(hit["_source"])
                                # print(context)

                            answers = self._reader_pool.model.answer_question(
                                params.question, context, stride=params.tokenstride)
                            for answer in answers:
                                answer["index"] = i
                                answer_holder.append(answer)

            # sort answers by probability
            answer_holder = sorted(
                answer_holder, key=lambda k: k['probability'], reverse=True)
            elapsed_time = time.time() - start_time
            response = {"answers": answer_holder, "took": elapsed_time}
            return response

        @router.post("/documents")
        async def get_documents(params: Document):
            """Get a list of documents and highlights that match the given search query

            Returns:
                dictionary -- contains details on elastic search results.
            """

            included_fields = ["name"]
            opinion_excerpt_length = 500

            # return only included fields + script_field,
            # limit response to top max_documents matches return highlights
            search_query = {
                "_source": included_fields,
                "query": {
                    "multi_match": {
                        "query":    params.question,
                        "fields": ["casebody.data.opinions.text", "name"]
                    }
                },
                "script_fields": {
                    "opinion_excerpt": {
                        "script": "(params['_source']['casebody']['data']['opinions'][0]['text']).substring(0," + str(opinion_excerpt_length) + ")"
                    }
                },
                "highlight": {
                    "fragment_size": params.highlight_span,
                    "fields": {
                        "casebody.data.opinions.text": {},
                        "name": {}
                    }
                },
                "size": params.max_documents
            }

            query_result = self._retriever.run_query(
                params.retriever, search_query)
            return query_result

        @router.post("/explain")
        async def get_explanation(params: Explanation):
            """Return  an explanation for a given model

            Returns:
                [dictionary]: [explanation , query, question, ]
            """

            context = params.context.replace(
                "<em>", "").replace("</em>", "")

            gradients, token_words, token_types, answer_text = self._reader_pool.model.explain_model(
                params.question, context)

            explanation_result = {"gradients": gradients,
                                  "token_words": token_words,
                                  "token_types": token_types,
                                  "answer": answer_text
                                  }
            return explanation_result



import click
from neuralqa.server import launch_server
from neuralqa.utils import cli_args
from neuralqa.utils import import_case_data, ConfigParser
import os


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
@cli_args.HOST
@cli_args.PORT
@cli_args.WORKERS
@cli_args.CONFIG_PATH
def test(host, port, workers, config_path):
    config = ConfigParser(config_path)
    print(config.config["reader"]["models"])
    # import_case_data()


@cli.command()
@cli_args.HOST
@cli_args.PORT
@cli_args.WORKERS
@cli_args.CONFIG_PATH
def ui(host, port, workers, config_path):
    if (config_path):
        os.environ["NEURALQA_CONFIG_PATH"] = config_path
    launch_server(host, port, workers)


if __name__ == '__main__':
    cli()
